import functools
import itertools
from datetime import timedelta

import singer
from singer import utils, get_bookmark, metadata
from singer.catalog import write_catalog, Catalog
from .client import Client
from .discover import discover
from .sync import sync_report

LOGGER = singer.get_logger()


# TODO: Add an integration test with multiple profiles that asserts state
def clean_state_for_report(config, state, tap_stream_id):
    top_level_bookmark = get_bookmark(state,
                                      tap_stream_id,
                                      'last_report_date')
    if top_level_bookmark:
        top_level_bookmark = utils.strptime_to_utc(top_level_bookmark)
        LOGGER.info("%s - Converting state to multi-profile format.", tap_stream_id)
        view_ids = get_view_ids(config)
        for view_id in view_ids:
            state = singer.write_bookmark(state,
                                          tap_stream_id,
                                          view_id,
                                          {'last_report_date': top_level_bookmark.strftime("%Y-%m-%d")})
        state = singer.clear_bookmark(state, tap_stream_id, 'last_report_date')
        singer.write_state(state)
    return state

def get_start_date(config, view_id, state, tap_stream_id):
    """
    Returns a date bookmark in state for the given stream, or the
    `start_date` from config, if no bookmark exists.
    """
    return utils.strptime_to_utc(get_bookmark(state,
                                              tap_stream_id,
                                              view_id,
                                              default={}).get('last_report_date',
                                                              config['start_date']))

def get_end_date(config):
    """
    Returns the end_date for the reporting sync. Under normal operation,
    this is defined as that date portion of UTC now.

    This can be overridden by the `end_date` config.json value.
    """
    if 'end_date' in config:
        return utils.strptime_to_utc(config['end_date'])
    return utils.now().replace(hour=0, minute=0, second=0, microsecond=0)

def get_view_ids(config):
    return config.get('view_ids') or [config.get('view_id')]

def do_sync(client, config, catalog, state):
    """
    Translate metadata into a set of metrics and dimensions and call out
    to sync to generate the required reports.
    """
    selected_streams = catalog.get_selected_streams(state)
    for stream in selected_streams:
        # Transform state for this report to new format before proceeding
        state = clean_state_for_report(config, state, stream.tap_stream_id)

        state = singer.set_currently_syncing(state, stream.tap_stream_id)
        singer.write_state(state)

        metrics = []
        dimensions = []
        mdata = metadata.to_map(stream.metadata)
        for field_path, field_mdata in mdata.items():
            if field_path == tuple():
                continue
            if field_mdata.get('inclusion') == 'unsupported':
                continue
            _, field_name = field_path
            if field_mdata.get('inclusion') == 'automatic' or \
               field_mdata.get('selected') or \
               (field_mdata.get('selected-by-default') and field_mdata.get('selected') is None):
                if field_mdata.get('behavior') == 'METRIC':
                    metrics.append(field_name)
                elif field_mdata.get('behavior') == 'DIMENSION':
                    dimensions.append(field_name)

        view_ids = get_view_ids(config)

        # NB: Resume from previous view for this report, dropping all
        # views before it to keep streams moving forward
        current_view = state.get('currently_syncing_view')
        if current_view:
            if current_view in view_ids:
                view_not_current = functools.partial(lambda cv, v: v != cv, current_view)
                view_ids = list(itertools.dropwhile(view_not_current, view_ids))
            else:
                state.pop('currently_syncing_view', None)

        reports_per_view = [{"profile_id": view_id,
                             "name": stream.stream,
                             "id": stream.tap_stream_id,
                             "metrics": metrics,
                             "dimensions": dimensions}
                            for view_id in view_ids]

        end_date = get_end_date(config)

        schema = stream.schema.to_dict()

        singer.write_schema(
            stream.stream,
            schema,
            stream.key_properties
            )

        for report in reports_per_view:
            state['currently_syncing_view'] = report['profile_id']
            singer.write_state(state)

            start_date = get_start_date(config, report['profile_id'], state, report['id'])

            sync_report(client, schema, report, start_date, end_date, state)
        state.pop('currently_syncing_view', None)
        singer.write_state(state)
    state = singer.set_currently_syncing(state, None)
    singer.write_state(state)

def do_discover(client, config):
    """
    Make request to discover.py and write result to stdout.
    """
    catalog = discover(client, config, get_view_ids(config))
    write_catalog(catalog)

def validate_config_view_ids(config):
    if 'view_id' not in config and 'view_ids' not in config:
        raise Exception("Config Validation Error: config.json MUST contain one of: view_ids, view_id.")
    if 'view_id' in config and 'view_ids' in config:
        raise Exception("Config Validation Error: config.json must ONLY contain view_id or view_ids, but not both.")

@utils.handle_top_exception(LOGGER)
def main():
    required_config_keys = ['start_date']
    args = singer.parse_args(required_config_keys)
    validate_config_view_ids(args.config)
    if "refresh_token" in args.config:  # if refresh_token in config assume OAuth2 credentials
        args.config['auth_method'] = "oauth2"
        additional_config_keys = ['client_id', 'client_secret', 'refresh_token']
    else:  # otherwise, assume Service Account details should be present
        args.config['auth_method'] = "service_account"
        additional_config_keys = ['client_email', 'private_key']

    singer.utils.check_config(args.config, additional_config_keys)

    config = args.config
    client = Client(config)
    catalog = args.catalog or Catalog([])
    state = args.state

    if args.properties and not args.catalog:
        raise Exception("DEPRECATED: Use of the 'properties' parameter is not supported. Please use --catalog instead")

    if args.discover:
        do_discover(client, config)
    else:
        do_sync(client, config, catalog, state)

if __name__ == "__main__":
    main()



from datetime import timedelta
import hashlib
import json
import singer
from singer import Transformer

LOGGER = singer.get_logger()

def generate_sdc_record_hash(raw_report, row, start_date, end_date):
    """
    Generates a SHA 256 hash to be used as the primary key for records
    associated with a report. This consists of a UTF-8 encoded JSON list
    containing:
    - The account_id, web_property_id, profile_id of the associated report
    - Pairs of ("ga:dimension_name", "dimension_value")
    - Report start_date value in YYYY-mm-dd format
    - Report end_date value in YYYY-mm-dd format

    Start and end date are included to maintain flexibility in the event the
    tap is extended to support wider date ranges.

    WARNING: Any change in the hashing mechanism, data, or sorting will
    REQUIRE a major version bump! As it will invalidate all previous
    primary keys and cause new data to be appended.
    """
    dimensions_headers = raw_report["reports"][0]["columnHeader"].get("dimensions", [])
    profile_id = raw_report["profileId"]
    web_property_id = raw_report["webPropertyId"]
    account_id = raw_report["accountId"]

    dimensions_pairs = sorted(zip(dimensions_headers, row.get("dimensions", [])), key=lambda x: x[0])

    # NB: Do not change the ordering of this list, it is the source of the PK hash
    hash_source_data = [account_id,
                        web_property_id,
                        profile_id,
                        dimensions_pairs,
                        start_date.strftime("%Y-%m-%d"),
                        end_date.strftime("%Y-%m-%d")]

    hash_source_bytes = json.dumps(hash_source_data).encode('utf-8')
    return hashlib.sha256(hash_source_bytes).hexdigest()


def generate_report_dates(start_date, end_date):
    total_days = (end_date - start_date).days
    # NB: Add a day to be inclusive of both start and end
    for day_offset in range(total_days + 1):
        yield start_date + timedelta(days=day_offset)

def report_to_records(raw_report):
    """
    Parse a single report object into Singer records, with added runtime info and PK.

    NOTE: This function assumes one report being run for one date range
    per request. For optimizations, the structure of the response will
    change, and this will need to be refactored.
    """
    # TODO: Handle data sampling keys and values, either in the records or as a separate stream? They look like arrays.
    # - https://developers.google.com/analytics/devguides/reporting/core/v4/rest/v4/reports/batchGet#ReportData
    report = raw_report["reports"][0]
    column_headers = report["columnHeader"]
    metrics_headers = [mh["name"] for mh in column_headers["metricHeader"]["metricHeaderEntries"]]
    dimensions_headers = column_headers.get("dimensions", [])

    for row in report.get("data", {}).get("rows", []):
        record = {}
        record.update(zip(dimensions_headers, row.get("dimensions", [])))
        record.update(zip(metrics_headers, row["metrics"][0]["values"]))

        report_date = raw_report["reportDate"]
        _sdc_record_hash = generate_sdc_record_hash(raw_report, row, report_date, report_date)
        record["_sdc_record_hash"] = _sdc_record_hash

        report_date_string = report_date.strftime("%Y-%m-%d")
        record["start_date"] = report_date_string
        record["end_date"] = report_date_string

        record["account_id"] = raw_report["accountId"]
        record["web_property_id"] = raw_report["webPropertyId"]
        record["profile_id"] = raw_report["profileId"]

        yield record

def sync_report(client, schema, report, start_date, end_date, state):
    """
    Run a sync, beginning from either the start_date or bookmarked date,
    requesting a report per day, until the last full day of data. (e.g.,
    "Yesterday")

    report = {"name": stream.tap_stream_id,
              "profile_id": view_id,
              "metrics": metrics,
              "dimensions": dimensions}
    """
    LOGGER.info("Syncing %s for view_id %s", report['name'], report['profile_id'])
    all_data_golden = True
    # TODO: Is it better to query by multiple days if `ga:date` is present?
    # - If so, we can optimize the calls here to generate date ranges and reduce request volume
    for report_date in generate_report_dates(start_date, end_date):
        for raw_report_response in client.get_report(report['name'], report['profile_id'],
                                                     report_date, report['metrics'],
                                                     report['dimensions']):

            with singer.metrics.record_counter(report['name']) as counter:
                time_extracted = singer.utils.now()
                with Transformer() as transformer:
                    for rec in report_to_records(raw_report_response):
                        singer.write_record(report["name"],
                                            transformer.transform(rec, schema),
                                            time_extracted=time_extracted)
                        counter.increment()

                # NB: Bookmark all days with "golden" data until you find the first non-golden day
                # - "golden" refers to data that will not change in future
                #   requests, so we can use it as a bookmark
                is_data_golden = raw_report_response["reports"][0]["data"].get("isDataGolden")

                if all_data_golden:
                    singer.write_bookmark(state,
                                          report["id"],
                                          report['profile_id'],
                                          {'last_report_date': report_date.strftime("%Y-%m-%d")})
                    singer.write_state(state)
                    if not is_data_golden:
                        # Stop bookmarking on first "isDataGolden": False
                        all_data_golden = False
    LOGGER.info("Done syncing %s for view_id %s", report['name'], report['profile_id'])



import re
from functools import reduce
import singer
from singer import metadata, Schema, CatalogEntry, Catalog

from tap_google_analytics.reports import PREMADE_REPORTS

LOGGER = singer.get_logger()

integer_field_overrides = {'ga:cohortNthDay',
                           'ga:cohortNthMonth',
                           'ga:cohortNthWeek',
                           'ga:daysSinceLastSession',
                           'ga:daysToTransaction',
                           'ga:nthDay',
                           'ga:nthHour',
                           'ga:nthMinute',
                           'ga:nthMonth',
                           'ga:nthWeek',
                           'ga:pageDepth',
                           'ga:screenDepth',
                           'ga:sessionCount',
                           'ga:sessionsToTransaction',
                           'ga:subContinentCode',
                           'ga:visitCount',
                           'ga:visitLength',
                           'ga:visitsToTransaction'}

datetime_field_overrides = {'ga:date',
                            'ga:dateHour'}

float_field_overrides = {'ga:latitude',
                         'ga:longitude',
                         'ga:avgScreenviewDuration',
                         'ga:avgSearchDuration',
                         'ga:avgSessionDuration',
                         'ga:avgTimeOnPage',
                         'ga:cohortSessionDurationPerUser',
                         'ga:cohortSessionDurationPerUserWithLifetimeCriteria',
                         'ga:searchDuration',
                         'ga:sessionDuration',
                         'ga:timeOnPage',
                         'ga:timeOnScreen'}

# pylint: disable=too-many-return-statements
def type_to_schema(ga_type, field_id):
    if field_id in datetime_field_overrides:
        return {"type": ["string", "null"], "format": "date-time"}
    elif ga_type == 'CURRENCY':
        # https://developers.google.com/analytics/devguides/collection/protocol/v1/parameters#ecomm
        return {"type": ["number", "null"]}
    elif ga_type == 'PERCENT':
        return {"type": ["number", "null"]}
    elif ga_type == 'TIME':
        return {"type": ["string", "null"]}
    elif ga_type == 'INTEGER' or field_id in integer_field_overrides:
        return {"type": ["integer", "null"]}
    elif ga_type == 'FLOAT' or field_id in float_field_overrides:
        return {"type": ["number", "null"]}
    elif ga_type == 'STRING':
        return {"type": ["string", "null"]}
    else:
        raise Exception("Unknown Google Analytics type: {}".format(ga_type))

def sort_schemas(sub_schemas):
    schemas = [None, None, None, None]
    for sub_schema in sub_schemas:
        if "integer" in sub_schema["type"]:
            schemas[0] = sub_schema
        elif "number" in sub_schema["type"]:
            schemas[1] = sub_schema
        elif sub_schema.get("format") == "date-time":
            schemas[2] = sub_schema
        elif "string" in sub_schema["type"]:
            schemas[3] = sub_schema
        else:
            raise Exception("Error sorting schemas, could not find slot for sub_schema: {}".format(sub_schema))
    return [s for s in schemas if s is not None]

def types_to_schema(ga_types, field_id):
    sub_schemas = [type_to_schema(t, field_id) for t in set(ga_types)]
    if len(sub_schemas) == 1:
        return sub_schemas[0]
    return {"anyOf": sort_schemas(sub_schemas)}

def is_static_XX_field(field_id, cubes_lookup):
    """
    GA has fields that are documented using a placeholder of `XX`, where
    the `XX` is replaced with a number in practice.

    Some of these are standard fields with constant numeric
    representations. These must be handled differently from other field,
    so this function will detect this case using the information we have
    gleaned in our cubes_lookup values.

    If the cubes_lookup map does NOT have the `XX` version in it, this
    function assumes that it has only the numeric versions.
    """
    return ('XX' in field_id
            and field_id not in cubes_lookup
            and field_id not in ["ga:metricXX", "ga:dimensionXX"])

def is_dynamic_XX_field(field_id, cubes_lookup):
    """
    GA has fields that are documented using a placeholder of `XX`, where
    the `XX` is replaced with a number in practice.

    Some of these are standard fields that are generated based on other
    artifacts defined for the profile (e.g., goals). These must be handled
    differently from other fields as well, since the IDs must be
    discovered through their own means.

    If the cubes_lookup map DOES have the `XX` version in it, this
    function assumes that the field is dynamically discovered.
    """
    return ('XX' in field_id
            and field_id in cubes_lookup
            and field_id not in ["ga:metricXX", "ga:dimensionXX"])

def handle_static_XX_field(field, cubes_lookup):
    """
    Uses a regex of the `XX` field's ID to discover which numeric versions
    of a given `XX` field name we have cubes for.

    Generates a schema entry and metadata for each.
    Returns:
    - Sub Schemas  {"<numeric_field_id>": {...field schema}, ...}
    - Sub Metadata {"numeric_field_id>": {...cubes metadata value}, ...}
    """
    regex_matcher = field['id'].replace("XX", r'\d\d?')
    matching_cubes = {field_id: cubes_lookup[field_id]
                      for field_id in cubes_lookup.keys()
                      if re.match(regex_matcher, field_id)}

    sub_schemas = {field_id: type_to_schema(field["dataType"], field["id"])
                   for field_id in matching_cubes.keys()}
    sub_metadata = matching_cubes

    return sub_schemas, sub_metadata


goal_related_field_ids = ['ga:goalXXStarts',
                          'ga:goalXXCompletions',
                          'ga:goalXXValue',
                          'ga:goalXXConversionRate',
                          'ga:goalXXAbandons',
                          'ga:goalXXAbandonRate',
                          'ga:searchGoalXXConversionRate']

def get_dynamic_fields_named(client, field, profile_id):
    """
    For known field types, retrieve their numeric forms through the client
    on a case-by-case basis (e.g., goals)
    """
    if field['id'] in goal_related_field_ids:
        return [{**field,
                 "id": field['id'].replace('XX', str(i)),
                 "profiles": [profile_id]}
                for i in client.get_goals_for_profile(profile_id)]
    else:
        # Skip unknown, or already handled, dynamic fields
        return []

def handle_dynamic_XX_field(client, field, cubes_lookup, profile_ids):
    """
    Discovers dynamic names of a given XX field using `client` with
    `get_dynamic_field_names` and matches them with the cubes known
    for the `XX` version of the name.

    Generates a schema entry and metadata for each.
    Returns:
    - Sub Schemas  {"<numeric_field_id>": {...field schema}, ...}
    - Sub Metadata {"numeric_field_id>": {...cubes metadata value}, ...}
    """
    # Do the logic of all profiles
    dynamic_field_names_per_profile = {}
    for profile_id in profile_ids:
        dynamic_field_names_per_profile[profile_id] = get_dynamic_fields_named(client, field, profile_id)

    dynamic_superfields = get_custom_fields_supertypes(dynamic_field_names_per_profile)

    sub_schemas = {d['id']: types_to_schema(d["dataTypes"], field["id"])
                   for d in dynamic_superfields}

    sub_metadata = {r['id']: cubes_lookup[field['id']]
                    for r in dynamic_superfields}

    dynamic_fields_support = calculate_custom_fields_support(dynamic_field_names_per_profile)

    return sub_schemas, sub_metadata, dynamic_fields_support

def write_metadata(mdata, field, cubes, custom_fields_support=None):
    """ Translate a field_info object and its cubes into its metadata, and write it. """
    unsupported_profiles = (custom_fields_support or {}).get(field['id'])
    if custom_fields_support is None or not unsupported_profiles:
        mdata = metadata.write(mdata, ("properties", field["id"]), "inclusion", "available")
    else:
        mdata = metadata.write(mdata, ("properties", field["id"]), "inclusion", "unsupported")
        mdata = metadata.write(mdata,
                               ("properties", field["id"]),
                               "unsupported-description",
                               "This field cannot be selected because it is not defined in profile(s): {}".format(
                                   ', '.join(unsupported_profiles)))
    mdata = metadata.write(mdata, ("properties", field["id"]), "tap_google_analytics.cubes", list(cubes))
    mdata = metadata.write(mdata, ("properties", field["id"]), "behavior", field["type"])
    mdata = metadata.write(mdata, ("properties", field["id"]), "tap_google_analytics.group", field["group"])

    return mdata

def generate_base_schema():
    return {"type": "object", "properties": {"_sdc_record_hash": {"type": "string"},
                                             "start_date": {"type": "string",
                                                            "format": "date-time"},
                                             "end_date": {"type": "string",
                                                          "format": "date-time"},
                                             "account_id": {"type": "string"},
                                             "web_property_id": {"type": "string"},
                                             "profile_id": {"type": "string"}}}

def generate_base_metadata(all_cubes, schema):
    mdata = metadata.get_standard_metadata(schema=schema, key_properties=["_sdc_record_hash"])
    mdata = metadata.to_map(mdata)
    mdata = metadata.write(mdata, (), "tap_google_analytics.all_cubes", list(all_cubes))
    mdata = reduce(lambda mdata, field_name: metadata.write(mdata, ("properties", field_name), "inclusion", "automatic"),
                   ["_sdc_record_hash", "start_date", "end_date", "account_id", "web_property_id", "profile_id"],
                   mdata)
    mdata = reduce(lambda mdata, field_name: metadata.write(mdata, ("properties", field_name), "tap_google_analytics.group", "Report Fields"),
                   ["_sdc_record_hash", "start_date", "end_date", "account_id", "web_property_id", "profile_id"],
                   mdata)
    return mdata

def calculate_custom_fields_support(custom_fields):
    """
    Return a dictionary of `{field_id: set(profiles_that_don't_support_it)}`.
    `custom_fields` should be of the form: {profile_id: [field_definition, ...], ...}
    """
    all_profiles = set(custom_fields.keys())
    custom_fields_sets = {}
    for fields in custom_fields.values():
        for field in fields:
            custom_fields_sets[field['id']] = (custom_fields_sets.get(field['id'], all_profiles) -
                                               set(field['profiles']))

    return custom_fields_sets

def get_custom_fields_supertypes(custom_fields):
    """
    Returns a list of fields with their aggregate things.
    E.g., [{
        "id": "ga:metric1",
        "kind": "analytics#customMetric",
        "dataTypes": ["string", "currency", ...],
        "type": "METRIC",
        "group": "Custom Fields"
    }, ...]
    """
    super_fields = {}
    for fields in custom_fields.values():
        for field in fields:
            super_field = super_fields.get(field['id'], {"id": field["id"],
                                                         "kind": field.get("kind"),
                                                         "dataTypes": set(),
                                                         "type": field["type"],
                                                         "group": field["group"]})
            super_field['dataTypes'].add(field['dataType'])
            super_fields[field['id']] = super_field
    return list(super_fields.values())

def generate_catalog_entry(client, standard_fields, custom_fields, all_cubes, cubes_lookup, profile_ids):
    schema = generate_base_schema()
    mdata = generate_base_metadata(all_cubes, schema)

    for standard_field in standard_fields:
        if (standard_field['status'] == 'DEPRECATED'
                or standard_field['id'] in ["ga:metricXX", "ga:dimensionXX"]):
            continue
        if is_static_XX_field(standard_field["id"], cubes_lookup):
            sub_schemas, sub_mdata = handle_static_XX_field(standard_field, cubes_lookup)
            schema["properties"].update(sub_schemas)
            for calculated_id, cubes in sub_mdata.items():
                specific_field = {**standard_field, **{"id": calculated_id}}
                mdata = write_metadata(mdata, specific_field, cubes)
        elif is_dynamic_XX_field(standard_field["id"], cubes_lookup):
            sub_schemas, sub_mdata, dynamic_fields_support = handle_dynamic_XX_field(client,
                                                                                     standard_field,
                                                                                     cubes_lookup,
                                                                                     profile_ids)
            schema["properties"].update(sub_schemas)
            for calculated_id, cubes in sub_mdata.items():
                specific_field = {**standard_field, **{"id": calculated_id}}
                mdata = write_metadata(mdata, specific_field, cubes, dynamic_fields_support)
        else:
            schema["properties"][standard_field["id"]] = type_to_schema(standard_field["dataType"],
                                                                        standard_field["id"])
            mdata = write_metadata(mdata,
                                   standard_field,
                                   cubes_lookup[standard_field["id"]])

    custom_fields_support = calculate_custom_fields_support(custom_fields)
    custom_super_fields = get_custom_fields_supertypes(custom_fields)
    for custom_field in custom_super_fields:
        if custom_field["kind"] == 'analytics#customDimension':
            cubes_lookup_name = 'ga:dimensionXX'
        elif custom_field["kind"] == 'analytics#customMetric':
            cubes_lookup_name = 'ga:metricXX'
        else:
            raise Exception('Unknown custom field "kind": {}'.format(custom_field["kind"]))

        cubes = cubes_lookup[cubes_lookup_name]

        mdata = write_metadata(mdata, custom_field, cubes, custom_fields_support)
        schema["properties"][custom_field["id"]] = types_to_schema(custom_field["dataTypes"],
                                                                   custom_field["id"])

    return schema, mdata

def generate_premade_catalog_entry(standard_fields, all_cubes, cubes_lookup):
    schema = generate_base_schema()
    mdata = generate_base_metadata(all_cubes, schema)

    for standard_field in standard_fields:
        # No dynamic fields in standard reports
        if (standard_field['status'] == 'DEPRECATED'
                or standard_field['id'] in ["ga:metricXX", "ga:dimensionXX"]):
            continue

        schema["properties"][standard_field["id"]] = type_to_schema(standard_field["dataType"],
                                                                    standard_field["id"])
        mdata = write_metadata(mdata, standard_field, cubes_lookup[standard_field["id"]])
    return schema, mdata

def generate_cubes_lookup(raw_cubes):
    """
    Generates a map of {field_id: cubes_list} for use in generating
    tap metadata for the catalog.
    """
    cubes_lookup = {}
    for raw_cube, fields in raw_cubes.items():
        for field in fields:
            if field not in cubes_lookup:
                cubes_lookup[field] = set()
            cubes_lookup[field].add(raw_cube)
    return cubes_lookup

def parse_cube_definitions(client):
    """
    Requests cube definitions from Google Metrics and Dimensions
    Explorer, and parses it into a structure for metadata usage.

    Returns:
       all_cubes -> names of all cubes that exist
       cubes_lookup -> mapping of field name to compatible cubes
    """
    raw_cubes = client.get_raw_cubes()
    all_cubes = set(raw_cubes.keys())
    cubes_lookup = generate_cubes_lookup(raw_cubes)
    return all_cubes, cubes_lookup

def get_custom_metrics(client, profile_id):
    custom_metrics = client.get_custom_metrics_for_profile(profile_id)
    metrics_fields = {"id", "name", "kind", "active", "min_value", "max_value"}
    account_id = client.profile_lookup[profile_id]["account_id"]
    web_property_id = client.profile_lookup[profile_id]["web_property_id"]
    profiles = client.get_profiles_for_property(account_id, web_property_id)
    return  [{"account_id": account_id,
              "web_property_id": web_property_id,
              "profiles": profiles,
              "type": "METRIC",
              "dataType": item["type"],
              "group": "Custom Variables or Columns",
              **{k:v for k,v in item.items() if k in metrics_fields}}
             for item in custom_metrics['items']]

def get_custom_dimensions(client, profile_id):
    custom_dimensions = client.get_custom_dimensions_for_profile(profile_id)
    dimensions_fields = {"id", "name", "kind", "active"}
    account_id = client.profile_lookup[profile_id]["account_id"]
    web_property_id = client.profile_lookup[profile_id]["web_property_id"]
    profiles = client.get_profiles_for_property(account_id, web_property_id)
    return [{"dataType": "STRING",
             "account_id": account_id,
             "web_property_id": web_property_id,
             "profiles": profiles,
             "type": "DIMENSION",
             "group": "Custom Variables or Columns",
             **{k:v for k,v in item.items() if k in dimensions_fields}}
            for item in custom_dimensions['items']]

def get_custom_fields(client, profile_id):
    custom_metrics_and_dimensions = []
    custom_metrics_and_dimensions.extend(get_custom_dimensions(client, profile_id))
    custom_metrics_and_dimensions.extend(get_custom_metrics(client, profile_id))
    return custom_metrics_and_dimensions


def transform_field(field):
    interesting_attributes = {k: v for k, v in field["attributes"].items()
                              if k in {"dataType", "group", "status", "type"}}
    return {"id": field["id"], "name": field["attributes"]["uiName"], **interesting_attributes}

def get_standard_fields(client):
    metadata_response = client.get_field_metadata()
    # NB: These fields' specific names aren't discoverable, we think
    #     "customVar*" is deprecated and "calcMetric" is beta.
    unsupported_fields = {"ga:customVarValueXX", "ga:customVarNameXX", "ga:calcMetric_<NAME>"}
    return [transform_field(f) for f in metadata_response["items"] if f["id"] not in unsupported_fields]

def generate_catalog(client, report_config, standard_fields, custom_fields, all_cubes, cubes_lookup, profile_ids):
    """
    Generate a catalog entry for each report specified in `report_config`
    """
    catalog_entries = []
    for report in PREMADE_REPORTS:
        metrics_dimensions = set(report['metrics'] + report['dimensions'])
        selected_by_default = {*report['metrics'][:10], # Use first 10 metrics in definition
                               *report.get('default_dimensions', [])}
        premade_fields = [field for field in standard_fields if field['id'] in metrics_dimensions]
        schema, mdata = generate_premade_catalog_entry(premade_fields,
                                                       all_cubes,
                                                       cubes_lookup)

        mdata = reduce(lambda mdata, field_name: metadata.write(mdata,
                                                                ("properties", field_name),
                                                                "selected-by-default", True),
                       selected_by_default,
                       mdata)

        catalog_entries.append(CatalogEntry(schema=Schema.from_dict(schema),
                                            key_properties=['_sdc_record_hash'],
                                            stream=report['name'],
                                            tap_stream_id=report['name'],
                                            metadata=metadata.to_list(mdata)))

    for report in report_config:
        schema, mdata = generate_catalog_entry(client,
                                               standard_fields,
                                               custom_fields,
                                               all_cubes,
                                               cubes_lookup,
                                               profile_ids)

        catalog_entries.append(CatalogEntry(schema=Schema.from_dict(schema),
                                            key_properties=['_sdc_record_hash'],
                                            stream=report['name'],
                                            tap_stream_id=report['id'],
                                            metadata=metadata.to_list(mdata)))
    return Catalog(catalog_entries)

def discover(client, config, profile_ids):
    # Draw from spike to discover all the things
    # Get field_infos (standard and custom)
    report_config = config.get("report_definitions") or []
    LOGGER.info("Discovering standard fields...")
    standard_fields = get_standard_fields(client)
    LOGGER.info("Discovering custom fields...")
    custom_fields = {}
    for profile_id in profile_ids:
        custom_fields[profile_id] = get_custom_fields(client, profile_id)
    LOGGER.info("Parsing cube definitions...")
    all_cubes, cubes_lookup = parse_cube_definitions(client)
    LOGGER.info("Generating catalog...")
    return generate_catalog(client, report_config, standard_fields, custom_fields, all_cubes, cubes_lookup, profile_ids)



#!/usr/bin/env python
from setuptools import setup

setup(
    name="tap-google-analytics",
    version="0.4.0",
    description="Singer.io tap for extracting data",
    author="Stitch",
    url="http://singer.io",
    classifiers=["Programming Language :: Python :: 3 :: Only"],
    py_modules=["tap_google_analytics"],
    install_requires=[
        "singer-python==5.9.0",
        "requests==2.22.0",
        "backoff==1.8.0",
        "jwt==0.6.1"
    ],
    extras_require={
        'dev': [
            'ipdb==0.11',
            'pylint',
            'nose'
        ]
    },
    entry_points="""
    [console_scripts]
    tap-google-analytics=tap_google_analytics:main
    """,
    packages=["tap_google_analytics"],
    package_data = {
        "tap_google_analytics": ["tap_google_analytics/ga_cubes.json"]
    },
    include_package_data=True,
)



from datetime import timedelta
import json
import math
import os
from jwt import (
    JWT,
    jwk_from_pem,
)
import requests
import singer
from singer import utils
import backoff

LOGGER = singer.get_logger()

def should_giveup(e):
    if e.response.status_code == 429:
        error_message = e.response.json().get("error", {}).get("message")
        if error_message:
            LOGGER.info("Encountered 429, backing off exponentially. Details: %s", error_message)
    return not e.response.status_code == 429

def _is_json(response):
    try:
        response.json()
        return True
    except Exception:
        return False


# pylint: disable=too-many-instance-attributes
class Client():
    def __init__(self, config):
        self.auth_method = config['auth_method']
        if self.auth_method == "oauth2":
            self.refresh_token = config["refresh_token"]
            self.client_id = config["client_id"]
            self.client_secret = config["client_secret"]
        elif self.auth_method == "service_account":
            self.client_email = config["client_email"]
            self.private_key = config["private_key"].encode()

        self.__access_token = None
        self.expires_in = 0
        self.last_refreshed = None

        self.quota_user = config.get("quota_user")
        self.user_agent = config.get("user_agent")

        self.session = requests.Session()
        if self.user_agent:
            self.session.headers.update({"User-Agent": self.user_agent})

        self.profile_lookup = {}
        self.__populate_profile_lookup()

    def __populate_profile_lookup(self):
        """
        Get all profiles available and associate them with their web property
        and account IDs to be looked up later during discovery.
        """
        for account_id in self.get_accounts_for_token():
            for web_property_id in self.get_web_properties_for_account(account_id):
                for profile_id in self.get_profiles_for_property(account_id, web_property_id):
                    self.profile_lookup[profile_id] = {"web_property_id": web_property_id,
                                                       "account_id": account_id}

    # Authentication and refresh
    def _ensure_access_token(self):
        if self.last_refreshed is not None and \
           (utils.now() - self.last_refreshed).total_seconds() < self.expires_in:
            return

        LOGGER.info("Refreshing access token.")
        self.last_refreshed = utils.now()

        if self.auth_method == "oauth2":
            payload = {
                "refresh_token": self.refresh_token,
                "client_id": self.client_id,
                "client_secret": self.client_secret,
                "grant_type": "refresh_token"
            }
        else:
            message = {
                "iss": self.client_email,
                "scope": "https://www.googleapis.com/auth/analytics.readonly",
                "aud":"https://oauth2.googleapis.com/token",
                "exp": math.floor((self.last_refreshed + timedelta(hours=1)).timestamp()),
                "iat": math.floor(self.last_refreshed.timestamp())
            }
            signing_key = jwk_from_pem(self.private_key)
            payload = {
                "grant_type": "urn:ietf:params:oauth:grant-type:jwt-bearer",
                "assertion": JWT().encode(message, signing_key, 'RS256')
            }

        token_response = requests.post("https://oauth2.googleapis.com/token", data=payload)

        token_response.raise_for_status()

        token_json = token_response.json()
        self.__access_token = token_json['access_token']
        self.expires_in = token_json['expires_in']


    @backoff.on_exception(backoff.expo,
                          (requests.exceptions.RequestException),
                          max_tries=10,
                          giveup=should_giveup,
                          factor=4,
                          jitter=None)
    def _make_request(self, method, url, params=None, data=None):
        params = params or {}
        data = data or {}

        self._ensure_access_token()

        headers = {"Authorization" : "Bearer " + self.__access_token}
        if self.quota_user:
            params["quotaUser"] = self.quota_user

        if method == 'POST':
            response = self.session.post(url, headers=headers, params=params, json=data)
        else:
            response = self.session.request(method, url, headers=headers, params=params)

        error_message = _is_json(response) and response.json().get("error", {}).get("message")
        if response.status_code == 400 and error_message:
            raise Exception("400 Client Error: Bad Request, details: {}".format(error_message))

        response.raise_for_status()

        return response

    def get(self, url, params=None):
        return self._make_request("GET", url, params=params)

    def post(self, url, data=None):
        return self._make_request("POST", url, data=data)

    # Discovery requests

    def get_field_metadata(self):
        metadata_response = self.get("https://www.googleapis.com/analytics/v3/metadata/{reportType}/columns".format(reportType="ga"))
        return metadata_response.json()

    def get_raw_cubes(self):
        try:
            cubes_response = self.get("https://ga-dev-tools.appspot.com/ga_cubes.json")
            cubes_response.raise_for_status()
            cubes_json = cubes_response.json()
        except Exception as ex:
            LOGGER.warning("Error fetching raw cubes, falling back to local copy. Exception message: %s", ex)
            local_cubes_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), "ga_cubes.json")
            with open(local_cubes_path, "r") as f:
                cubes_json = json.load(f)
        return cubes_json

    def get_accounts_for_token(self):
        """ Return a list of account IDs available to hte associated token. """
        accounts_response = self.get('https://www.googleapis.com/analytics/v3/management/accounts')
        account_ids = [i['id'] for i in accounts_response.json()['items']]
        return account_ids

    def get_web_properties_for_account(self, account_id):
        """ Return a list of webproperty IDs for the account specified. """
        webprops_response = self.get('https://www.googleapis.com/analytics/v3/management/accounts/{accountId}/webproperties'.format(accountId=account_id))
        webprops_ids = [w['id'] for w in webprops_response.json()['items']]
        return webprops_ids

    def get_profiles_for_property(self, account_id, web_property_id):
        """
        Gets all profiles for property to associate with custom metrics and dimensions.
        """
        profiles_url = 'https://www.googleapis.com/analytics/v3/management/accounts/{accountId}/webproperties/{webPropertyId}/profiles'
        profiles_response = self.get(profiles_url.format(accountId=account_id,
                                                         webPropertyId=web_property_id))
        return [p["id"] for p in profiles_response.json()['items']]

    def get_goals_for_profile(self, profile_id):
        """
        Gets all goals for a profile_id.
        """
        return self.get_goals(self.profile_lookup[profile_id]["account_id"],
                              self.profile_lookup[profile_id]["web_property_id"],
                              profile_id)

    def get_goals(self, account_id, web_property_id, profile_id):
        """
        Gets all goal IDs for property and account to name custom metrics and dimensions.
        """
        profiles_url = 'https://www.googleapis.com/analytics/v3/management/accounts/{accountId}/webproperties/{webPropertyId}/profiles/{profileId}/goals'
        goals_response = self.get(profiles_url.format(accountId=account_id,
                                                      webPropertyId=web_property_id,
                                                      profileId=profile_id))
        return [g["id"] for g in goals_response.json()['items']]

    def get_custom_metrics_for_profile(self, profile_id):
        """
        Get all custom metrics associated with the given profile ID.
        """
        return self.get_custom_metrics(self.profile_lookup[profile_id]["account_id"],
                                       self.profile_lookup[profile_id]["web_property_id"])

    def get_custom_metrics(self, account_id, web_property_id):
        """
        Gets all metrics for the specified web_property_id.

        """
        metrics_url = 'https://www.googleapis.com/analytics/v3/management/accounts/{accountId}/webproperties/{webPropertyId}/customMetrics'

        custom_metrics_response = self.get(metrics_url.format(accountId=account_id,
                                                              webPropertyId=web_property_id))
        return custom_metrics_response.json()

    def get_custom_dimensions_for_profile(self, profile_id):
        """
        Get all custom dimensions associated with the given profile ID.
        """
        return self.get_custom_dimensions(self.profile_lookup[profile_id]["account_id"],
                                          self.profile_lookup[profile_id]["web_property_id"])

    def get_custom_dimensions(self, account_id, web_property_id):
        """
        Gets all dimensions for the specified web_property_id
        """
        url = 'https://www.googleapis.com/analytics/v3/management/accounts/{accountId}/webproperties/{webPropertyId}/customDimensions'
        custom_dimensions_response = self.get(url.format(accountId=account_id,
                                                         webPropertyId=web_property_id))

        # NOTE: Assuming that all custom dimensions are STRING, since there's no type information
        return custom_dimensions_response.json()

    # Sync Requests w/ Pagination and token refresh
    # Docs for more info: https://developers.google.com/analytics/devguides/reporting/core/v4/rest/v4/reports/batchGet
    def get_report(self, name, profile_id, report_date, metrics, dimensions):
        """
        Parameters:
        - name - the tap_stream_id of the report being run
        - profile_id - the profile for which this report is being run
        - report_date - the day to retrieve data for, as a Python datetime object, to limit report data
        - metrics - list of metrics, of the form ["ga:metric1", "ga:metric2", ...]
        - dimensions - list of dimensions, of the form ["ga:dim1", "ga:dim2", ...]

        Returns:
        - A generator of a sequence of reports w/ associated metadata (metrics/dims/report_date/profile)
        """
        nextPageToken = None
        # TODO: Optimization, if speed is an issue, up to 5 requests can be placed per HTTP batch
        # - This will require changes to all parsing code to account for multiple report responses coming back
        while True:
            report_date_string = report_date.strftime("%Y-%m-%d")
            LOGGER.info("Making report request for profile ID %s and date %s (nextPageToken: %s)",
                        profile_id,
                        report_date_string,
                        nextPageToken)
            body = {"reportRequests":
                    [{"viewId": profile_id,
                      "dateRanges": [{"startDate": report_date_string,
                                      "endDate": report_date_string}],
                      "metrics": [{"expression": m} for m in metrics],
                      "dimensions": [{"name": d} for d in dimensions]}]}
            if nextPageToken:
                body["reportRequests"][0]["pageToken"] = nextPageToken
            with singer.metrics.http_request_timer(name):
                report_response = self.post("https://analyticsreporting.googleapis.com/v4/reports:batchGet", body)
            report = report_response.json()

            # Assoc in the request data to be used by the caller
            report.update({"profileId": profile_id,
                           "webPropertyId": self.profile_lookup[profile_id]["web_property_id"],
                           "accountId": self.profile_lookup[profile_id]["account_id"],
                           "reportDate": report_date,
                           "metrics": metrics,
                           "dimensions": dimensions})

            yield report

            # NB: Assumes only one report at a time
            nextPageToken = report["reports"][0].get("nextPageToken")

            if not nextPageToken:
                break



PREMADE_REPORTS = [
    {
        "name": "Audience Overview",
        "metrics": [
            "ga:users",
            "ga:newUsers",
            "ga:sessions",
            "ga:sessionsPerUser",
            "ga:pageviews",
            "ga:pageviewsPerSession",
            "ga:avgSessionDuration",
            "ga:bounceRate"
        ],
        "dimensions": [
            "ga:date",
            "ga:language",
            "ga:country",
            "ga:city",
            "ga:browser",
            "ga:operatingSystem",
            "ga:screnResolution",
            "ga:year",
            "ga:month",
            "ga:hour",
        ],
        "default_dimensions": [
            "ga:date"
        ]
    },
    {
        "name": "Audience Geo Location",
        "metrics": [
            "ga:users",
            "ga:newUsers",
            "ga:sessions",
            "ga:pageviewsPerSession",
            "ga:avgSessionDuration",
            "ga:bounceRate"
        ],
        "dimensions": [
            "ga:date",
            "ga:year",
            "ga:month",
            "ga:hour",
            "ga:country",
            "ga:city",
            "ga:continent",
            "ga:subContinent"
        ],
        "default_dimensions": [
            "ga:date",
            "ga:country",
            "ga:city",
            "ga:continent",
            "ga:subContinent"
        ]
    },
    {
        "name": "Audience Technology",
        "metrics": [
            "ga:users",
            "ga:newUsers",
            "ga:sessions",
            "ga:pageviewsPerSession",
            "ga:avgSessionDuration",
            "ga:bounceRate"
        ],
        "dimensions": [
            "ga:date",
            "ga:year",
            "ga:month",
            "ga:hour",
            "ga:browser",
            "ga:operatingSystem",
            "ga:screenResolution",
            "ga:screenColors",
            "ga:flashVersion",
            "ga:javaEnabled",
            "ga:hostname"
        ],
        "default_dimensions": [
            "ga:date",
            "ga:browser",
            "ga:operatingSystem",
        ]
    },
    {
        "name": "Acquisition Overview",
        "metrics": [
            "ga:sessions",
            "ga:pageviewsPerSession",
            "ga:avgSessionDuration",
            "ga:bounceRate"
        ],
        "dimensions": [
            'ga:acquisitionMedium',
            'ga:acquisitionSource',
            'ga:acquisitionSourceMedium',
            'ga:acquisitionTrafficChannel'
        ],
        "default_dimensions": [
            "ga:acquisitionTrafficChannel",
            "ga:acquisitionSource",
            "ga:acquisitionSourceMedium",
            "ga:acquisitionMedium",
        ]
    },
    {
        "name": "Behavior Overview",
        "metrics": [
            "ga:pageviews",
            "ga:uniquePageviews",
            "ga:avgTimeOnPage",
            "ga:bounceRate",
            "ga:exitRate",
            "ga:exits"
        ],
        "dimensions": [
            "ga:date",
            "ga:year",
            "ga:month",
            "ga:hour",
            "ga:pagePath",
            "ga:pageTitle",
            "ga:searchKeyword",
            "ga:eventCategory"
        ],
        "default_dimensions": [
            "ga:date",
            "ga:pagePath",
            "ga:pageTitle",
            "ga:searchKeyword"
        ]
    },
    {
        "name": "Ecommerce Overview",
        "metrics": ["ga:transactions"],
        "dimensions": [
            "ga:transactionId",
            "ga:campaign",
            "ga:source",
            "ga:medium",
            "ga:keyword",
            "ga:socialNetwork"
        ],
        "default_dimensions": [
            "ga:transactionId",
            "ga:campaign",
            "ga:source",
            "ga:medium",
            "ga:keyword",
            "ga:socialNetwork"
        ]
    }
]
